"""Init

Revision ID: 7ccb2156db5e
Revises: 
Create Date: 2025-05-19 10:18:42.388542

"""
from typing import Sequence, Union

# revision identifiers, used by Alembic.
revision: str = '7ccb2156db5e'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###
